// Copyright (C) 2013 You SUZUKI
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
// ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <stdio.h>
#include <stdint.h>
#include <vector>
#include <map>

#define LOG(...) fprintf(stderr, __VA_ARGS__)
#define ERR	(fprintf(stderr, "%s(%d): ERR\n", __FILE__, __LINE__))

typedef uint32_t DWORD;
typedef unsigned int UINT;
typedef uint16_t USHORT;

namespace ttf {

int getword(FILE *fp)
{
	if (fp == NULL)
		return ERR, -1;
	int c = fgetc(fp);
	if (c == EOF)
		return -1;	// EOF
	int d = fgetc(fp);
	if (d == EOF)
		return ERR, -1;
	return (c << 8) | d;
}

template <typename T>
int getdword(FILE *fp, T &r)
{
	if (fp == NULL)
		return ERR, -1;
	int c = fgetc(fp);
	if (c == EOF)
		return -1;	// EOF
	int d = fgetc(fp);
	int e = fgetc(fp);
	int f = fgetc(fp);
	if ((d == EOF) || (e == EOF) || (f == EOF))
		return ERR, -1;
	r = (c << 24U) | (d << 16) | (e << 8) | f;
	return 0;
}

template <typename T>
DWORD id(const T *s)
{
	if (s[0] == 0)
		return 0;
	if (s[1] == 0)
		return s[0] << 24U;
	if (s[2] == 0)
		return (s[0] << 24U) | (s[1] << 16);
	if (s[3] == 0)
		return (s[0] << 24U) | (s[1] << 16) | (s[2] << 8);
	return (s[0] << 24U) | (s[1] << 16) | (s[2] << 8) | s[3];
}

void putprintablechar(int c)
{
	if ((c < 0x20) || (c >= 0x7f))
		LOG("\\x%02X", c);
	else
		LOG("%c", c);
}

void putid(DWORD dw)
{
	putprintablechar((dw >> 24) & 0xff);
	putprintablechar((dw >> 16) & 0xff);
	putprintablechar((dw >> 8) & 0xff);
	putprintablechar(dw & 0xff);
}

}

class SeekAndResume {
  public:
	SeekAndResume(FILE *fp, int off) {
		origpos = ftell(mfp = fp);
		fseek(mfp, off, SEEK_SET);
	}
	~SeekAndResume() {
		fseek(mfp, origpos, SEEK_SET);
	}
	FILE *mfp;
	int origpos;
};

int getversion(FILE *fp)
{
	DWORD ver;
	if (ttf::getdword(fp, ver) < 0)
		return ERR -1;
	LOG("  version: %u.%04x\n", ver >> 16, ver&0xffff);
	return 0;
}

int getoffset(FILE *fp)
{
	int us = ttf::getword(fp);
	if (us < 0) return ERR, -1;
	LOG("offset=%d(%04x)\n", us, us);
	return us;
}

int maxp(FILE *fp, DWORD off)
{
	SeekAndResume sr(fp, off);
	if (getversion(fp) < 0)
		return ERR, -1;
	int us = ttf::getword(fp);
	if (us < 0)
		return ERR, -1;
	LOG("  number of glyphs: %d\n", us);
	return 0;
}

int scripttable(FILE *fp, int off, std::vector<int> &reqlst)
{
	SeekAndResume sr(fp, off);
	LOG("    DefaultLangSys "); getoffset(fp);
	int cnt = ttf::getword(fp);
	if (cnt < 0) return ERR, -1;
	LOG("    LangSysCount: %d\n", cnt);
	for (int i = 0; i < cnt; i++) {
		LOG("    Lang %d: ", i);
		DWORD id;
		if (ttf::getdword(fp, id) < 0)
			return ERR, -1;
		LOG("Tag="); ttf::putid(id); LOG(", ");
		int langsys = getoffset(fp);
		if (langsys < 0) return ERR, -1;
		if (id == ttf::id("JAN ")) {
			SeekAndResume sr(fp, off + langsys + 2);
			int s = ttf::getword(fp);
			if (s < 0) return ERR, -1;
			LOG("      ReqFeatureIndex: %d\n", s);
			int cnt = ttf::getword(fp);
			if (cnt < 0) return ERR, -1;
			LOG("      FeatureCount: %d\n", cnt);
			for (int i = 0; i < cnt; i++) {
				LOG("      Index %d: ", i);
				int f = ttf::getword(fp);
				if (f < 0) return ERR, -1;
				LOG("%d\n", f);
				reqlst.push_back(f);
			}
			return 1;
		}
	}
	return 0;
}

int scriptlist(FILE *fp, int off, std::vector<int> &reqlst)
{
	SeekAndResume sr(fp, off);
	int cnt = ttf::getword(fp);
	if (cnt < 0) return ERR, -1;
	LOG("  Count: %d\n", cnt);
	for (int i = 0; i < cnt; i++) {
		DWORD id;
		int offscr;
		LOG("  Script %d: ", i);
		if (ttf::getdword(fp, id) < 0) return ERR, -1;
		LOG("Tag="); ttf::putid(id); LOG(", ");
		offscr = getoffset(fp);
		if (offscr < 0) return ERR, -1;
		if (id == ttf::id("kana")) {
			int rc = scripttable(fp, off + offscr, reqlst);
			if (rc < 0) return ERR, -1;
			if (rc > 0) return rc;
		}
	}
	return 0;
}

struct FEATURE {
	DWORD id;
	int off;
	FEATURE(DWORD i, int o) { id = i; off = o; }
};

int featurelist(FILE *fp, int off, std::vector<FEATURE> &flst)
{
	SeekAndResume sr(fp, off);
	int cnt = ttf::getword(fp);
	LOG("  FeatureList: count %d\n", cnt);
	for (int i = 0; i < cnt; i++) {
		DWORD id;
		if (ttf::getdword(fp, id) < 0) return ERR, -1;
		int offset = ttf::getword(fp);
		if (offset < 0) return ERR, -1;
		if (id == ttf::id("vert")) {
			LOG("    %d: ", i);
			ttf::putid(id);
			LOG(" offset=%d\n", offset);
		}
		flst.push_back(FEATURE(id, off + offset));
	}
	return 0;
}

int coverage(FILE *fp, int off, std::vector<int> &covlst)
{
	SeekAndResume sr(fp, off);
	int fmt = ttf::getword(fp);
	if (fmt < 0) return ERR, -1;
	LOG("format=%d, ", fmt);

	int cnt;
	switch (fmt) {
	  case 1:
		cnt = ttf::getword(fp);
		if (cnt < 0) return ERR, -1;
		LOG("glyphcount: %d\n", cnt);
		for (int i = 0; i < cnt; i++) {
			int id = ttf::getword(fp);
			if (id < 0) return ERR, -1;
			covlst.push_back(id);
		}
		break;
	  default:
		return -1;
//	  case 2:
//		printf("\t    RangeCount: %d\n", cnt = ttf::getword(fp));
//		for (int i = 0; i < cnt; i++) {
//			printf("\t      %d: start=%d, end=%d, startCoverageIndex=%d\n",
//				i, ttf::getword(fp), ttf::getword(fp), ttf::getword(fp));
//		}
//		break;
	}
	return 0;
}

int subtable(FILE *fp, int off)
{
	std::vector<int> covlst;
	SeekAndResume sr(fp, off);
	int substfmt = ttf::getword(fp);
	int cov, cnt;
	if (substfmt < 0) return ERR, -1;
	LOG("      substfmt=%d, ", substfmt);
	switch (substfmt) {
	  case 2:
		cov = ttf::getword(fp);
		if (cov < 0) return ERR, -1;
		LOG("coverage=%d, ", cov);
		if (coverage(fp, off + cov, covlst) < 0) return ERR, -1;
		cnt = ttf::getword(fp);
		if (cnt < 0) return ERR, -1;
		LOG("      subst-glyphcount=%d ... ", cnt);
		printf("  %d beginsubsttable\n", cnt);
		for (int i = 0; i < cnt; i++) {
			int id = ttf::getword(fp);
			if (id < 0) return ERR, -1;
			printf("  @ %d %d ;\n", covlst[i], id);
		}
		printf("  endsubsttable\n");
		LOG("done.\n");
		break;
	  default:
		return ERR, -1;
	}
	return 0;
}

int lookuplist(FILE *fp, int off, std::vector<int> &lulst)
{
	SeekAndResume sr(fp, off);
	int cnt;
	LOG("  LookupList: count=%d\n", cnt = ttf::getword(fp));
	for (int i = 0; i < cnt; i++) {
		int toff = ttf::getword(fp);
		if (toff < 0) return ERR, -1;
		lulst.push_back(off + toff);
	}
	return 0;
}

int lookuptable(FILE *fp, int off)
{
	SeekAndResume sr(fp, off);
	int lutype = ttf::getword(fp);
	if (lutype < 0) return ERR, -1;
	LOG("type=%d, ", lutype);
	int flag = ttf::getword(fp);
	if (flag < 0) return ERR, -1;
	LOG("flag=%04X, ", flag);
	int cnt = ttf::getword(fp);
	if (cnt < 0) return ERR, -1;
	LOG("count=%d\n", cnt);
	if (lutype != 1) return ERR, -1;
	for (int i = 0; i < cnt; i++) {
		LOG("    subtable %d: ", i);
		int subtbl = getoffset(fp);
		if (subtbl < 0) return ERR, -1;
		if (subtable(fp, off + subtbl) < 0) return ERR, -1;
	}
	int mark = -1;
	if (flag & 0x10) {
		mark = ttf::getword(fp);
		LOG("    MarkFilteringSet=%d\n", mark);
	}
	return 0;
}

int featuretable(FILE *fp, int off, std::vector<int> &lutbl)
{
	SeekAndResume sr(fp, off);
	LOG("params ");
	int p = ttf::getword(fp);
	if (p < 0) return ERR, -1;
	if (p == 0) LOG("none"); else LOG("%d", p);

	int cnt = ttf::getword(fp);
	if (cnt < 0) return ERR, -1;
	LOG(", cnt=%d\n", cnt);
	for (int i = 0; i < cnt; i++) {
		int luidx = ttf::getword(fp);
		if (luidx < 0) return ERR, -1;
		LOG("    %d: LookupIndex=%d\n", i, luidx);
		lutbl.push_back(luidx);
	}
	return 0;
}

int gsub(FILE *fp, DWORD off)
{
	int rc = 0;
	std::vector<int> reqlst, lutbl, lulst;
	std::vector<FEATURE> flst;

	SeekAndResume sr(fp, off);
	if (getversion(fp) < 0)
		return ERR, -1;

	LOG("  ScriptList "); int scrl = getoffset(fp);
	if (scrl < 0) return ERR, -1;
	rc = scriptlist(fp, off + scrl, reqlst);
	if (rc < 0) return ERR, -1;
	if (rc == 0) return 0;

	LOG("  FeatureList ");
	int ftrl = getoffset(fp);
	if (ftrl < 0) return ERR, -1;
	if (featurelist(fp, off + ftrl, flst) < 0) return ERR, -1;

	std::vector<int>::iterator i;
	rc = 0;
	for (i = reqlst.begin(); i != reqlst.end(); i++) {
		if (flst[*i].id == ttf::id("vert")) {
			LOG("  Feature %d: ", *i);
			if (featuretable(fp, flst[*i].off, lutbl) < 0)
				return ERR, -1;
			rc = 1;
		}
	}
	if (rc == 0) return 0;

	LOG("  LookupList ");
	int lul = getoffset(fp);
	if (lul < 0) return ERR, -1;
	if (lookuplist(fp, off + lul, lulst) < 0) return ERR, -1;
	for (i = lutbl.begin(); i != lutbl.end(); i++) {
		LOG("    LookupTable %d: position=%d, ", *i, lulst[*i]);
		if (lookuptable(fp, lulst[*i]) < 0)
			return ERR, -1;
	}
	return 1;
}

int main(int argc, char *argv[])
{
	if (argc < 2)
		return ERR, 1;
	FILE *fp = fopen(argv[1], "rb");
	if (fp == NULL)
		return ERR, 2;
	DWORD id;
	if (ttf::getdword(fp, id) < 0)
		return ERR, 2;
	if (id != ttf::id("ttcf"))
		return ERR, 3;
	DWORD ver;
	if (ttf::getdword(fp, ver) < 0)
		return ERR, 4;
	if (ver != 0x10000)
		return ERR, 5;
	int numFonts;
	if (ttf::getdword(fp, numFonts) < 0)
		return ERR, 6;
	LOG("number of fonts: %d\n", numFonts);

	for (int i = 0; i < numFonts; i++) {
		LOG("font #%d: ", i);
		printf("%d subfont\n", i);
		DWORD dw;
		if (ttf::getdword(fp, dw) < 0)
			return ERR, 7;
		LOG("offset table starts from: %u (0x%x)\n", (UINT)dw, (UINT)dw);
		SeekAndResume sr(fp, dw);
		if (ttf::getdword(fp, ver) < 0)
			return ERR, 8;
		if (ver != 0x10000)
			return ERR, 9;
		int numTables = ttf::getword(fp);
		if (numTables < 0)
			return ERR, 10;
		LOG("number of tables: %d\n", numTables);
		if (ttf::getword(fp) < 0)	// search range
			return ERR, 11;
		if (ttf::getword(fp) < 0)	// entry selector
			return ERR, 11;
		if (ttf::getword(fp) < 0)	// range shift
			return ERR, 11;
		for (int i = 0; i < numTables; i++) {
			DWORD tag, sum, offset, len;
			if (ttf::getdword(fp, tag) < 0)
				return ERR, 12;
			if (ttf::getdword(fp, sum) < 0)
				return ERR, 13;
			if (ttf::getdword(fp, offset) < 0)
				return ERR, 13;
			if (ttf::getdword(fp, len) < 0)
				return ERR, 13;
			ttf::putid(tag);
			LOG(": offset %u (0x%x), length %u (0x%x)\n",
				(UINT)offset, (UINT)offset, (UINT)len, (UINT)len);
			if (tag == ttf::id("maxp")) {
				if (maxp(fp, offset) < 0)
					return ERR, 14;
			}
			else if (tag == ttf::id("GSUB")) {
				if (gsub(fp, offset) < 0)
					return ERR, 15;
			}
		}
		LOG("\n");
		printf("endsubfont\n");
	}

	fclose(fp);
	return 0;
}
